#!/usr/bin/python3
from PySide2 import QtCore, QtGui, QtWidgets
from playsound import playsound
from ckMseed.LibTrace import *
from getopt import getopt
from operator import mod
from glob import glob
import os
import sys

VERSION = "2022.1.0.0"

# import threading
SPACE = " "

# vars used for batchmode
BATCHMODE = 0

def Usage(err=''):
    print(err)
    print("Usage\nckMseed \nckMseed -#\nckMseed -h\nckMseed [-d DataDirs] "
          "[-s] [-v]")
    print("       -d DataDirs - colon separated list of data directories "
          "[default: cwd]")
    print("       -h Usage")
    print("       -s Simple mode. Check and report endianess of first fixed "
          "header only")
    print("       -v Check and report endianess of first fixed header and "
          "determine")
    print("           start times for first and last block of mseed file")
    print("       -V Read all header blockettes")
    print("       NOTE: -s, -v, & -V are mutually exclusive.-V supercedes -v "
          "supercedes -s")


class MainWindow(QtWidgets.QWidget):

    def __init__(self, BatchFile, Dataloc, RunFlag):
        super().__init__()

        self.WindowSetup()
        self.DataInit(BatchFile, Dataloc, RunFlag)

###############################################################################################
        
    def WindowSetup(self):
        """
        Build main window
        """
        # WINDOW SETTINGS

        self.setObjectName("Window")
        self.setFixedSize(730, 546) # static window size
        self.setWindowTitle("ckMseed " + VERSION)

        ###############################################################

        # WINDOW ASSETS

        # Frame around global assets
        self.frame = QtWidgets.QFrame(self)
        self.frame.setGeometry(QtCore.QRect(-2, 475, 750, 41))
        self.frame.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        # Save Scan button
        self.save_scan_btn = QtWidgets.QPushButton(self)
        self.save_scan_btn.setGeometry(QtCore.QRect(561, 483, 100, 25))
        self.save_scan_btn.setStyleSheet("QPushButton::hover"
                                         "{"
                                         "background-color:green;"
                                         "}")
        self.save_scan_btn.setText("Save Scan")
        self.save_scan_btn.clicked.connect(self.clicked_save_scan)

        # Exit button
        self.exit_btn = QtWidgets.QPushButton(self)
        self.exit_btn.setGeometry(QtCore.QRect(660, 483, 60, 25))
        self.exit_btn.setStyleSheet("QPushButton::hover"
                                    "{"
                                    "background-color:rgb(165, 29, 45);"
                                    "}")
        self.exit_btn.setText("Exit")
        self.exit_btn.clicked.connect(self.click_exit)

        # PopUp Help Check box
        self.pop_up_help = QtWidgets.QCheckBox(self)
        self.pop_up_help.setGeometry(QtCore.QRect(10, 483, 101, 23))
        self.pop_up_help.setChecked(False)
        self.pop_up_help.setText("PopUp Help")
        self.pop_up_help.setToolTip("Toggles 'PopUp Help' on and off")
        self.pop_up_help.toggled.connect(self.click_pop_up_help)

        # Banner 0, 522, 700, 20
        self.files_alert = QtWidgets.QTextBrowser(self)
        self.files_alert.setGeometry(QtCore.QRect(0, 522, 730, 24))
        self.files_alert.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        font = QtGui.QFont()
        font.setFamily("DejaVu Sans")
        self.files_alert.setFont(font)
        self.files_alert.setStyleSheet("background-color:yellow")

        # Tab setup
        self.tabWidget = QtWidgets.QTabWidget(self)
        self.tabWidget.setGeometry(QtCore.QRect(4, 10, 721, 460))

        ###############################################################
                
        # SCAN TAB
        self.Scan = QtWidgets.QWidget()
        self.Scan.setObjectName("Scan")

        # Frame around 'Data Directories'
        self.frame_1 = QtWidgets.QFrame(self.Scan)
        self.frame_1.setGeometry(QtCore.QRect(-2, 3, 721, 41))
        self.frame_1.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame_1.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        # Data Directories label
        self.dd_label = QtWidgets.QLabel(self.Scan)
        self.dd_label.setGeometry(QtCore.QRect(8, 15, 151, 17))
        self.dd_label.setText("Data Directories:")

        # Data Directories text
        self.dd_text = QtWidgets.QTextEdit(self.Scan)
        self.dd_text.setGeometry(QtCore.QRect(123, 11, 368, 25))
        self.dd_text.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.dd_text.setAcceptRichText(False)

        # Clear button
        self.clear_btn = QtWidgets.QPushButton(self.Scan)
        self.clear_btn.setGeometry(QtCore.QRect(650, 11, 60, 25))
        self.clear_btn.setStyleSheet("QPushButton::hover"
                                     "{"
                                     "background-color:yellow;"
                                     "}")
        self.clear_btn.setText("Clear")
        self.clear_btn.clicked.connect(self.click_clear_dd)

        # Find button
        self.find_btn = QtWidgets.QPushButton(self.Scan)
        self.find_btn.setGeometry(QtCore.QRect(596, 11, 55, 25))
        self.find_btn.setStyleSheet("QPushButton::hover"
                                     "{"
                                     "background-color:green;"
                                     "}")
        self.find_btn.setText("Find")
        self.find_btn.clicked.connect(self.file_dialogue)

        # Scan Traces button
        self.scan_traces_btn = QtWidgets.QPushButton(self.Scan)
        self.scan_traces_btn.setGeometry(QtCore.QRect(497, 11, 100, 25))
        self.scan_traces_btn.setStyleSheet("""
                                           QPushButton{
                                           background-color:rgb(98, 160, 234);;
                                           }
                                           QPushButton::hover
                                           {
                                           background-color:green;
                                           }
                                           """)
        self.scan_traces_btn.setText("Scan Traces")
        self.scan_traces_btn.clicked.connect(self.click_scan_traces)

        # Frame around Stations
        self.frame_2 = QtWidgets.QFrame(self.Scan)
        self.frame_2.setGeometry(QtCore.QRect(-2, 49, 721, 41))
        self.frame_2.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        # Stations label
        self.stations_label = QtWidgets.QLabel(self.Scan)
        self.stations_label.setGeometry(QtCore.QRect(8, 61, 265, 17))
        self.stations_label.setText("Scan only stations (colon separated list):")

        # Stations text
        self.stations_text = QtWidgets.QTextEdit(self.Scan)
        self.stations_text.setGeometry(QtCore.QRect(278, 57, 366, 25))
        self.stations_text.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.stations_text.setAcceptRichText(False)

        # Stations Clear button
        self.clear2_btn = QtWidgets.QPushButton(self.Scan)
        self.clear2_btn.setGeometry(QtCore.QRect(650, 57, 60, 25))
        self.clear2_btn.setStyleSheet("QPushButton::hover"
                                     "{"
                                     "background-color:yellow;"
                                     "}")
        self.clear2_btn.setText("Clear")
        self.clear2_btn.clicked.connect(self.click_clear_stations)

        # Frame around 'Scan Type'
        self.frame_3 = QtWidgets.QFrame(self.Scan)
        self.frame_3.setGeometry(QtCore.QRect(-2, 95, 721, 41))
        self.frame_3.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        # Scan Type label
        self.scan_type_label = QtWidgets.QLabel(self.Scan)
        self.scan_type_label.setGeometry(QtCore.QRect(8, 106, 100, 17))
        self.scan_type_label.setText("Scan Type:")

        # Simple button
        self.simple_rbtn = QtWidgets.QRadioButton(self.Scan)
        self.simple_rbtn.setGeometry(QtCore.QRect(85, 105, 70, 23))
        self.simple_rbtn.setChecked(False)
        self.simple_rbtn.clicked.connect(self.click_scan_type) # if clicked

        self.simple_label = QtWidgets.QLabel(self.Scan)
        self.simple_label.setGeometry(QtCore.QRect(105, 107, 45, 17))
        self.simple_label.setText("Simple")

        # Verbose button
        self.verbose_rbtn = QtWidgets.QRadioButton(self.Scan)
        self.verbose_rbtn.setGeometry(QtCore.QRect(160, 105, 70, 23))
        self.verbose_rbtn.setChecked(False)
        self.verbose_rbtn.clicked.connect(self.click_scan_type) # if clicked

        self.verbose_label = QtWidgets.QLabel(self.Scan)
        self.verbose_label.setGeometry(QtCore.QRect(180, 107, 55, 17))
        self.verbose_label.setText("Verbose")

        # Very verbose button
        self.vv_rbtn = QtWidgets.QRadioButton(self.Scan)
        self.vv_rbtn.setGeometry(QtCore.QRect(245, 105, 70, 23))
        self.vv_rbtn.setChecked(False)
        self.vv_rbtn.clicked.connect(self.click_scan_type) # if clicked 

        self.vv_label = QtWidgets.QLabel(self.Scan)
        self.vv_label.setGeometry(QtCore.QRect(265, 107, 100, 17))
        self.vv_label.setText("Very Verbose")

        
        # Stat:Chan:Loc:Net:Sps Found label
        self.ScanText_label = QtWidgets.QLabel(self.Scan)
        self.ScanText_label.setGeometry(QtCore.QRect(248, 150, 265, 17))
        self.ScanText_label.setText("Stat:Chan:Loc:Net:Sps Found")

        # frame around Stat:Chan:Loc:Net:Sps Found
        self.frame_4 = QtWidgets.QFrame(self.Scan)
        self.frame_4.setGeometry(QtCore.QRect(-2, 141, 730, 291))
        self.frame_4.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        # Stat:Chan:Loc:Net:Sps Found text
        self.ScanText = QtWidgets.QTextEdit(self.Scan)
        self.ScanText.setGeometry(QtCore.QRect(8, 175, 700, 225))
        self.ScanText.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.ScanText.setAcceptRichText(False)

        self.tabWidget.addTab(self.Scan, "Scan") # end of assets in tab

        ###############################################################

        # BIG ENDIAN TAB
        self.BigEndian = QtWidgets.QWidget()
        self.BigEndian.setObjectName("BigEndian")

        # 'Display Traces Matching' label
        self.display_traces_matching = QtWidgets.QLabel(self.BigEndian)
        self.display_traces_matching.setGeometry(QtCore.QRect(210, 10, 265, 17))
        self.display_traces_matching.setText("Display Traces Matching:")
        self.display_traces_matching.hide()

        # 'Display Traces Matching' drop menu
        self.big_e_dropmenu = QtWidgets.QComboBox(self.BigEndian)
        self.big_e_dropmenu.setGeometry(QtCore.QRect(380, 10, 100, 17))
        self.big_e_dropmenu.activated.connect(self.big_e_dropmenu_selections)
        self.big_e_dropmenu.hide()

        # Big Endian Textbox
        self.big_e_text = QtWidgets.QTextEdit(self.BigEndian)
        self.big_e_text.setGeometry(QtCore.QRect(8, 38, 700, 355))
        self.big_e_text.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.big_e_text.setAcceptRichText(False)
        self.big_e_text.hide()


        self.tabWidget.addTab(self.BigEndian, "Big Endian") # end of assets in tab
        self.BigEndian.hide()

        ###############################################################

        # LITTLE ENDIAN TAB
        self.LittleEndian = QtWidgets.QWidget()
        self.LittleEndian.setObjectName("LittleEndian")

        # 'Display Traces Matching' label
        self.display_traces_matching2 = QtWidgets.QLabel(self.LittleEndian)
        self.display_traces_matching2.setGeometry(QtCore.QRect(210, 10, 265, 17))
        self.display_traces_matching2.setText("Display Traces Matching:")
        self.display_traces_matching2.hide()

        # 'Display Traces Matching' drop menu
        self.little_e_dropmenu = QtWidgets.QComboBox(self.LittleEndian)
        self.little_e_dropmenu.setGeometry(QtCore.QRect(380, 10, 100, 17))
        self.little_e_dropmenu.activated.connect(self.little_e_dropmenu_selections)
        self.little_e_dropmenu.hide()

        # Little Endian Textbox
        self.little_e_text = QtWidgets.QTextEdit(self.LittleEndian)
        self.little_e_text.setGeometry(QtCore.QRect(8, 38, 700, 355))
        self.little_e_text.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.little_e_text.setAcceptRichText(False)
        self.little_e_text.hide()


        self.tabWidget.addTab(self.LittleEndian, "Little Endian") # end of assets in tab
        self.LittleEndian.hide()

        ###############################################################

        # ERRORS TAB
        self.Errors = QtWidgets.QWidget()
        self.Errors.setObjectName("Errors")

        # Errors Textbox
        self.error_text = QtWidgets.QTextEdit(self.Errors)
        self.error_text.setGeometry(QtCore.QRect(8, 48, 700, 378))
        self.error_text.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.error_text.setAcceptRichText(False)
        self.error_text.hide()

        # Frame around 'Display' and radio buttons
        self.frame_5 = QtWidgets.QFrame(self.Errors)
        self.frame_5.setGeometry(QtCore.QRect(-2, 3, 721, 41))
        self.frame_5.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.frame_5.hide()

        # Display label
        self.display_label = QtWidgets.QLabel(self.Errors)
        self.display_label.setGeometry(QtCore.QRect(8, 15, 100, 17))
        self.display_label.setText("Display:")
        self.display_label.setToolTip("")
        self.display_label.hide()

        # All button
        self.all_rbtn = QtWidgets.QRadioButton(self.Errors)
        self.all_rbtn.setGeometry(QtCore.QRect(70, 13, 70, 23))
        self.all_rbtn.setChecked(True)
        self.all_rbtn.setText("All")
        self.all_rbtn.hide()

        # Read/Write button
        self.rw_rbtn = QtWidgets.QRadioButton(self.Errors)
        self.rw_rbtn.setGeometry(QtCore.QRect(120, 13, 100, 23))
        self.rw_rbtn.setChecked(False)
        self.rw_rbtn.setText("Read/Write")
        self.rw_rbtn.hide()

        # Size button
        self.size_rbtn = QtWidgets.QRadioButton(self.Errors)
        self.size_rbtn.setGeometry(QtCore.QRect(223, 13, 70, 23))
        self.size_rbtn.setChecked(False)
        self.size_rbtn.setText("Size")
        self.size_rbtn.hide()

        # Endian button
        self.endian_rbtn = QtWidgets.QRadioButton(self.Errors)
        self.endian_rbtn.setGeometry(QtCore.QRect(280, 13, 70, 23))
        self.endian_rbtn.setChecked(False)
        self.endian_rbtn.setText("Endian")
        self.endian_rbtn.hide()

        # Non-Unique button
        self.non_u_rbtn = QtWidgets.QRadioButton(self.Errors)
        self.non_u_rbtn.setGeometry(QtCore.QRect(355, 13, 100, 23))
        self.non_u_rbtn.setChecked(False)
        self.non_u_rbtn.setText("Non-Unique")
        self.non_u_rbtn.hide()

        self.tabWidget.addTab(self.Errors, "Errors") # end of assets in tab

        self.show()

###############################################################################################

    def DataInit(self, BatchFile, Dataloc, RunFlag):
        """
        Initialize variables and data and check if batchmode
        """
        # set timezone to UTC
        os.environ['TZ'] = 'GMT'
        time.tzname = ('GMT', 'GMT')
        time.timezone = 0

        # Variables
        self.BatchFile = BatchFile
        self.Dataloc = Dataloc
        self.DataDirs = ""
        self.StatSel = ""
        self.Savefile = ""
        self.NumBigFiles = 0
        self.NumLittleFiles = 0
        self.NumErrors = 0

        # Lists 
        self.StatSelList = []
        self.DataDirList = []
        self.StatSelList = []
        self.ErrorDict = []
        self.ErrorAll = []
        self.ErrorRW = []
        self.ErrorSize = []
        self.ErrorEndian = []
        self.ErrorUnique = []

        # Dictionaries 
        self.BigEndianDict = {}
        self.LittleEndianDict = {}
        self.ErrorDict = {}
        self.StatChanLocDict = {}

        if self.Dataloc:
            self.dd_text.setText(self.Dataloc)
        else:
            self.dd_text.setText(os.getcwd())

        # If batchmode then launch from here and exit
        if BATCHMODE:
            # address endian change first since unique i.e. no batchfile
            print("Running in Batch Mode")
            if RunEndianBatch:
                print("Changing Endianess to: ", Endianess)
                self.ToEndian = Endianes
                self.RunChangeEndian = 1
                print("Finding Files beneath directory: ", self.DataDirs)
                self.BuildTrcList()
                self.ChangeEndian()
            else:
                if RunTimeBatch:
                    print("Correcting Timing")
                if RunHeaderBatch:
                    print("Correcting Headers")

                print("Using Template: " + self.BatchFile)
                self.LoadTemplate()
                if RunTimeBatch:
                    if not self.UpdateTimeDict:
                        print("No Timing Corrections in Template.\nDone")
                        sys.exit(1)
                    print("Finding Files beneath directory: "
                          + self.DataDirs)
                    self.BuildTrcList()
                    self.RunApplyTimeCor = 1
                    self.ApplyTimeCor()
                if RunHeaderBatch:
                    if not self.UpdateHdrDict:
                        print("No Header Corrections in Template.\nDone")
                        sys.exit(1)
                    print("Finding Files beneath directory: " +
                          self.DataDirs)
                    self.BuildTrcList()
                    self.RunModHdrs = 1
                    self.ModHdrs()
            print("Done")
            sys.exit(0)

###############################################################################################

    def click_exit(self):
        """
        Quit program
        """
        quit()

###############################################################################################

    def click_clear_dd(self):
        """
        Clear directories text box
        """
        self.dd_text.clear()

###############################################################################################

    def click_clear_stations(self):
        """
        Clear stations text box
        """
        self.stations_text.clear()

###############################################################################################

    def click_scan_type(self):
        """
        Initialize ScanType
        """
        if self.simple_rbtn.isChecked():
            self.ScanType = 0
        elif self.verbose_rbtn.isChecked():
            self.ScanType = 1
        elif self.vv_rbtn.isChecked():
            self.ScanType = 2

###############################################################################################

    def click_errors_display(self):
        """
        Initialize ErrorType
        """
        if self.all_rbtn.isChecked():
            self.ErrorType = 0
        elif self.rw_rbtn.isChecked():
            self.ErrorType = 1
        elif self.size_rbtn.isChecked():
            self.ErrorType = 2
        elif self.endian_rbtn.isChecked():
            self.ErrorType = 3
        elif self.non_u_rbtn.isChecked():
            self.ErrorType = 4

###############################################################################################

    def click_pop_up_help(self):
        """
        Initialize tool tips on/off
        """
        if self.pop_up_help.isChecked():
            self.exit_btn.setToolTip("Exit Program")
            self.save_scan_btn.setToolTip("Allows scan results to\n"
                                          "be saved to a file")
            self.clear_btn.setToolTip("Clears 'Data Directories' entry")
            self.find_btn.setToolTip("Dialogue window to\n" 
                                     "select Data Directories")
            self.scan_traces_btn.setToolTip("Scan traces using\n"
                                            "'Data Directories' as\n"
                                            "top level directories")
            self.dd_label.setToolTip("Search path(s) for finding mseed files.\n"
                                     "Colon separate multiple entries")
            self.clear2_btn.setToolTip("Clear station filter")
            self.scan_type_label.setToolTip("Simple - ID traces\n"
                                            "Verbose - Include Start/End Times\n"
                                            "Very verbose - Scan all blocks")
            self.display_label.setToolTip("Filter Error messages")
        
        else:
            self.exit_btn.setToolTip("")
            self.save_scan_btn.setToolTip("")
            self.clear_btn.setToolTip("")
            self.find_btn.setToolTip("")
            self.scan_traces_btn.setToolTip("")
            self.dd_label.setToolTip("")
            self.clear2_btn.setToolTip("")
            self.scan_type_label.setToolTip("")
            self.display_label.setToolTip("")

###############################################################################################

    def file_dialogue(self):
        """
        File Dialogue for choosing directories
        """
        self.search = QtWidgets.QFileDialog.getExistingDirectory()

        directories = self.dd_text.toPlainText()

        # if data directories field is empty, add chosen directory
        if directories == "":
            if (self.search != ""): # if user doesn't cancel chosing a directory 
                self.dd_text.insertPlainText(self.search)               
        # if data directories field is not empty, append chosen directory with :
        else:
            if (self.search != ""):
                self.dd_text.moveCursor(QtGui.QTextCursor.End) # move cursor to end of entry
                self.dd_text.insertPlainText(":"+ self.search) # append

###############################################################################################

    def click_scan_traces(self):
        """
        Check scan type and at least one directory have been selected
        before BuildTrcList
        """
        directory = self.dd_text.toPlainText() # gather directory from 'Data Directories'

        # if data directory is empty and no scan type button has been selected, an error message will pop up
        if directory == "" and not self.simple_rbtn.isChecked() and not self.verbose_rbtn.isChecked() and not self.vv_rbtn.isChecked():
            error_msg3 = QtWidgets.QMessageBox()
            error_msg3.setIcon(QtWidgets.QMessageBox.Critical)
            error_msg3.setText("Error")
            error_msg3.setInformativeText("No directory chosen and no scan type selected.")
            error_msg3.setWindowTitle("Error")
            error_msg3.exec_()

        # if data directory is not empty but no scan type button has been selected, an error message will pop up
        if directory != "" and not self.simple_rbtn.isChecked() and not self.verbose_rbtn.isChecked() and not self.vv_rbtn.isChecked():
            error_msg2 = QtWidgets.QMessageBox()
            error_msg2.setIcon(QtWidgets.QMessageBox.Critical)
            error_msg2.setText("Error")
            error_msg2.setInformativeText("No scan type selected.")
            error_msg2.setWindowTitle("Error")
            error_msg2.exec_()

        # if data directory is empty but a scan type button has been selected, an error message will pop up
        if (directory == "") and (self.simple_rbtn.isChecked() or self.verbose_rbtn.isChecked() or self.vv_rbtn.isChecked()):
            error_msg4 = QtWidgets.QMessageBox()
            error_msg4.setIcon(QtWidgets.QMessageBox.Critical)
            error_msg4.setText("Error")
            error_msg4.setInformativeText("No directory chosen.")
            error_msg4.setWindowTitle("Error")
            error_msg4.exec_()

        # if data directory is not empty and a scan type button has been selected, continue
        if (directory != "") and (self.simple_rbtn.isChecked() or self.verbose_rbtn.isChecked() or self.vv_rbtn.isChecked()):

            self.ResetData() # reset data fields and lists/dicts if 
                             # this isn't the first time clicking scan traces

            # reveal tab contents and add options to drop menu
            self.display_traces_matching.show()
            self.big_e_text.show()
            self.display_traces_matching2.show()
            self.little_e_text.show()
            self.big_e_dropmenu.show()
            self.little_e_dropmenu.show()
            self.error_text.show()
            self.frame_5.show()
            self.display_label.show()
            self.all_rbtn.show()
            self.rw_rbtn.show()
            self.size_rbtn.show()
            self.endian_rbtn.show()
            self.non_u_rbtn.show()

            self.BuildTrcList()

###############################################################################################

    def BuildTrcList(self):
        """
        Build a trace list using DataDirList as base directories
        """

        directory = self.dd_text.toPlainText()

        # begin directory search
        for idir in directory.split(":"):
            dirlist = glob(idir)
            for newdir in dirlist:
                if not os.path.isdir(newdir):
                    err = "***WARNING*** Directory " + newdir + " not found."
                    self.addTextInfoBar(err, "red")
                    self.NumErrors = self.NumErrors + 1
                    return
                self.DataDirList.append(newdir)

        self.StatSel = self.stations_text.toPlainText()
        if self.StatSel != "":
            for statsel in self.StatSel.split(":"):
                #statsel = string.lstrip(statsel)
                self.StatSelList.append(statsel)

        self.RunScan = 1

        self.LaunchFindTrace()

        numfiles = self.NumBigFiles + self.NumLittleFiles
        numerrors = self.NumErrors
        self.UpdateInfoBar(numfiles, numerrors) # update info

        self.WriteScan()
        self.WriteBigEndian("All")
        self.WriteLittleEndian("All")
        #self.DisplayError()
        self.PlaySound()

###############################################################################################

    def ResetData(self):
        """
        Delete data from past scans 
        """

        # Variables 
        if self.NumBigFiles != 0:
            self.NumBigFiles = 0
        if self.NumLittleFiles != 0:
            self.NumLittleFiles = 0
        if self.NumErrors != 0:
            self.NumErrors = 0

        # Text fields
        data_found = self.ScanText.toPlainText()
        big_e_text = self.big_e_text.toPlainText()
        little_e_text = self.little_e_text.toPlainText()
        error_text = self.error_text.toPlainText()
        if data_found != "":
            self.ScanText.clear()
        if big_e_text != "":
            self.big_e_text.clear()
        if little_e_text != "":
            self.little_e_text.clear()
        if error_text != "":
            self.error_text.clear()
        self.big_e_dropmenu.clear()
        self.little_e_dropmenu.clear()

        # Lists
        if len(self.DataDirList) != 0:
            self.DataDirList.clear()
        if len(self.StatChanLocDict) != 0:
            self.StatChanLocDict.clear()
        if len(self.StatSelList) != 0:
            self.StatSelList.clear()
        if len(self.ErrorDict) != 0:
            self.ErrorDict.clear()
        if len(self.ErrorAll) != 0:
            self.ErrorAll.clear()
        if len(self.ErrorRW) != 0:
            self.ErrorRW.clear()
        if len(self.ErrorSize) != 0:
            self.ErrorSize.clear()
        if len(self.ErrorEndian) != 0:
            self.ErrorEndian.clear()
        if len(self.ErrorUnique) != 0:
            self.ErrorUnique.clear()

        # Dicts
        if len(self.BigEndianDict) != 0:
            self.BigEndianDict.clear()
        if len(self.LittleEndianDict) != 0:
            self.LittleEndianDict.clear()
        if len(self.ErrorDict) != 0:
            self.ErrorDict.clear()
        if len(self.StatChanLocDict) != 0:
            self.StatChanLocDict.clear()

###############################################################################################

    def AddToDict(self, inDict, id, inDir, inList):
        """
        Function to add data to dictionaries 
        """
        try:
            try:
                inDict[id][inDir].append(inList)
            except KeyError:
                inDict[id][inDir] = []
                inDict[id][inDir].append(inList)
        except KeyError:
            inDict[id] = {}
            inDict[id][inDir] = []
            inDict[id][inDir].append(inList)
        return

###############################################################################################

    def wait(self, words, cnt):
        """
        Puts words plus cnt to info bar
        """
        txt = words + str(cnt)
        self.ScanText.append(txt)

###############################################################################################

    def PlaySound(self):
        """
        Plays a sound for each scan trace
        """
        sound_file_name = "click.wav"
        playsound(sound_file_name)

###############################################################################################

    def WriteError(self, info, type="", color="black"):
        """
        Writes errors to appropriate list
        """
        if BATCHMODE:
            if info:
                print(info)
        else:
            info = info + "\n"
            self.ErrorAll.append((info, color))

            if type == "Read/Write":
                self.ErrorRW.append((info, color))
            elif type == "Size":
                self.ErrorSize.append((info, color))
            elif type == "Endian":
                self.ErrorEndian.append((info, color))
            elif type == "Unique":
                self.ErrorUnique.append((info, color))
            else:
                pass
        return

###############################################################################################

    def ReadAll(self, msfile, numblocks, inDir, inFile, old_fileid):
        """
        Attempts to read all blockettes and check for consistancy
        """
        # local variables
        # lstrip=string.strip
        # ljoin=string.join
        lfixedhdr = msfile.fixedhdr
        ltypenxt = msfile.typenxt
        lGetBlk = msfile.GetBlk
        lrate = msfile.rate
        lstat = msfile.FH.Stat.strip().decode("utf-8")
        lchan = msfile.FH.Chan.strip().decode("utf-8")
        lloc = msfile.FH.Loc.strip().decode("utf-8")
        lnet = msfile.FH.Net.strip().decode("utf-8")
        fullname = os.path.join(inDir, inFile)
        warn = ""
        # blk_err = 0
        blocks = []
        n = 0
        blksize = msfile.blksize
        while n < numblocks:
            # read & rewrite fixed header
            hdrs = lfixedhdr(n * blksize)
            rate = lrate
            stat = lstat
            chan = lchan
            loc = lloc
            net = lnet
            numblk = hdrs[3][3]
            addseek = hdrs[3][6]
            # fileid = ljoin(list(map(str,(stat,chan,loc,net,rate))), ":")
            templist = list(map(str, (stat, chan, loc, net, rate)))
            fileid = ":".join(templist)
            if fileid != old_fileid:
                warn = "WARNING: Multiplexed miniseed file"
                warn1 = "\tFile: " + fullname
            b = 0
            # loop over number of blockettes following fixed header at block n
            while b < numblk:
                seekval = (n * blksize) + addseek
                (blktype, next) = ltypenxt(seekval)
                blklist = lGetBlk(blktype, seekval)
                if not blklist:
                    # if blockette error encountered bail
                    err = "Corrupt mseed: " + fullname
                    self.WriteError(err, "Endian", "red")
                    err1 = "\tRecord: " + str(n)
                    self.WriteError(err1, "Endian")
                    err2 = "\tUnrecognized Blockette: " + str(blktype)
                    self.AddToDict(self.ErrorDict, "Unrecognized Blockette",
                                   inDir, inFile)
                    self.WriteError(err2, "Endian")
                    return None
                else:
                    if blktype not in blocks:
                        blocks.append(blktype)
                addseek = next
                b += 1
            old_fileid = fileid
            n += 1
        if warn:
            self.WriteError(warn, "Unique", "red")
            self.WriteError(warn1, "Unique")
            self.AddToDict(self.ErrorDict, "Multiplexed", inDir, inFile)
            self.NumErrors.set(self.NumErrors.get() + 1)

        return blocks

###############################################################################################

    def LaunchFindTrace(self):
        """
        Separated out so FindTrace could be run as a thread
        """
        #if not BATCHMODE:
        #    self.CancelTL(self.RunScan, "Scan Traces")

        self.Find_Trace(self.DataDirList)

        #if self.RunScan and not BATCHMODE:
        #    self.KillCancelTL(self.RunScan)

###############################################################################################

    def Find_Trace(self, DataDir):
        """
        Scan mseed files and fill lists with data
        """
        # local variables
        lstrftime = time.strftime
        lgmtime = time.gmtime

        stack = []
        for k in range(len(DataDir)):
            if not DataDir[k] in stack:
                stack.append(DataDir[k])

        cnt = 1

        while stack:
            directory = stack.pop()
            if not os.access(directory, 5):
                text = "Skipped: %s" % directory
                text1 = "\t\tAccess Error"
                self.WriteError(text, "Read/Write", "red")
                self.WriteError(text1, "Read/Write", "black")
                self.NumErrors = self.NumErrors + 1
                continue

            try:
                listfiles = (dir for dir in os.listdir(directory))
            except Exception as e:
                print("Directory Read Error: %s" % e)
                self.NumErrors = self.NumErrors + 1
                continue

            while 1:
                try:
                    file = next(listfiles)
                except StopIteration:
                    break

                # respond to a cancel request
                #if not self.RunScan.:
                #    break
                # keep user posted of progress
                if mod(cnt, 10):
                    pass
                else:
                    self.wait("Examining File: ", cnt)
                # build up list of mseed files w/ full pathnames
                fullname = os.path.join(directory, file)
                if os.path.isfile(fullname):
                    if not os.access(fullname, 6):
                        err = "ERROR: Read/Write permission denied."
                        err1 = "\t File:" + fullname
                        self.WriteError(err, "Read/Write", "red")
                        self.WriteError(err1, "Read/Write")
                        self.AddToDict(self.ErrorDict, "Permission Error",
                                       directory, file)
                        self.NumErrors = self.NumErrors + 1
                        continue

                    try:
                        msfile = Mseed(fullname)
                        if msfile.isMseed():
                            try:
                                # simple test to determine if correct size file
                                numblocks = msfile.numblocks
                            except Exception:
                                err = ("ERROR: Cannot determine file and "
                                       "block sizes.")
                                err1 = "\t File:" + fullname
                                self.WriteError(err, "Size", "red")
                                self.WriteError(err1, "Size")
                                self.AddToDict(
                                    self.ErrorDict, "Size", directory, file)
                                self.NumErrors = self.NumErrors + 1
                                continue
                            
                            # assign header info
                            rate = msfile.rate
                            stat = msfile.FH.Stat.strip().decode("utf-8")
                            chan = msfile.FH.Chan.strip().decode("utf-8")
                            loc = msfile.FH.Loc.strip().decode("utf-8")
                            net = msfile.FH.Net.strip().decode("utf-8")
                            templist = list(map(str, (stat, chan, loc, net,
                                                      rate)))
                            fileid = ":".join(templist)  

                            # window on specific stations if exist
                            if self.StatSelList:
                                if stat not in self.StatSelList:
                                    continue
                            if not self.ScanType:
                                filelist = [file]
                            else:
                                (startepoch, endepoch) = msfile.FirstLastTime()
                                endepoch += 1
                                start = lstrftime('%Y:%j:%H:%M:%S',
                                                  lgmtime(startepoch))
                                end = lstrftime('%Y:%j:%H:%M:%S',
                                                lgmtime(endepoch))
                                if self.ScanType == 1:
                                    filelist = [file, start, end]
                                else:
                                    blklist = self.ReadAll(msfile,
                                                           numblocks,
                                                           directory,
                                                           file,
                                                           fileid)
                                    filelist = [file, start, end] + blklist

                            if not self.ScanType:
                                self.StatChanLocDict[fileid] = []
                            else:
                                try:
                                    if startepoch < self.StatChanLocDict[fileid][0][0]:  # noqa: E501
                                        self.StatChanLocDict[fileid][0][0] =\
                                            startepoch
                                    if endepoch > self.StatChanLocDict[fileid][0][1]:  # noqa: E501
                                        self.StatChanLocDict[fileid][0][1] =\
                                            endepoch
                                except KeyError:
                                    self.StatChanLocDict[fileid] = []
                                    self.StatChanLocDict[fileid].append([startepoch, endepoch])

                            # build endian lists
                            if msfile.byteorder == "big":
                                self.NumBigFiles = self.NumBigFiles + 1
                                self.AddToDict(self.BigEndianDict, fileid,
                                               directory, filelist)
                            elif msfile.byteorder == "little":
                                self.NumLittleFiles = self.NumLittleFiles + 1
                                self.AddToDict(self.LittleEndianDict, fileid,
                                               directory, filelist)
                        msfile.close()
                    except Exception as e:
                        err = "ERROR: %s" % e
                        self.WriteError(err, "Read/Write", "red")
                        err1 = "\t File:" + fullname
                        self.WriteError(err1, "Read/Write")
                        self.AddToDict(
                            self.ErrorDict, "Read/Write", directory, file)
                        self.NumErrors = self.NumErrors + 1
                cnt += 1

                # add fullname to stack if it is a directory or link
                if os.path.isdir(fullname) or (os.path.islink(fullname) and
                                               not os.path.isfile(fullname)):
                    if fullname not in stack:
                        stack.append(fullname)

        # Add 'All' option to drop menu and set as default choice
        self.big_e_dropmenu.insertItem(0, "All")
        self.big_e_dropmenu.setCurrentIndex(0)
        for indexkey in list(self.BigEndianDict):
            self.big_e_dropmenu.addItem(indexkey)

        self.little_e_dropmenu.insertItem(0, "All")
        self.little_e_dropmenu.setCurrentIndex(0)
        for indexkey in list(self.LittleEndianDict):
            self.little_e_dropmenu.addItem(indexkey)

        return

###############################################################################################

    def WriteScan(self):
        """
        Add data found in file scan to Scan tab
        """
        Yellow = QtGui.QColor(255,255,0)
        White = QtGui.QColor(255,255,255)
        LightBlue = QtGui.QColor(153,204,255)

        scan_header = ("Stat:Chan:Loc:Net:Sps \t Big \t Little \t Start Time \t\t End Time")
        self.ScanText.setTextBackgroundColor(Yellow)
        self.ScanText.setText(scan_header)
        self.ScanText.setTextBackgroundColor(White)

        if not self.StatChanLocDict:
            return

        keylist = []
        keylist = list(self.StatChanLocDict.keys())
        keylist.sort()

        if keylist:
            c = 0
            for key in keylist:
                numBig = 0
                numLittle = 0

                if key in self.BigEndianDict:
                    for dir in list(self.BigEndianDict[key].keys()):
                        numBig = numBig + len(self.BigEndianDict[key][dir])
                if key in self.LittleEndianDict:
                    for dir in list(self.LittleEndianDict[key].keys()):
                        numLittle = numLittle + \
                            len(self.LittleEndianDict[key][dir])
               
                text = key + "\t" + str(numBig) + "\t" + str(numLittle)
                
                if not self.ScanType:
                    pass
                else:
                    start = time.strftime('%Y:%j:%H:%M:%S', time.gmtime(
                        self.StatChanLocDict[key][0][0]))
                    end = time.strftime('%Y:%j:%H:%M:%S', time.gmtime(
                        self.StatChanLocDict[key][0][1]))
                    templist = list(map(str, (start, end)))
                    text = (text + "\t" +
                            "   ".join(templist))  # python 3 compatible

                if c:
                    self.ScanText.setTextBackgroundColor(LightBlue)
                    self.ScanText.append(text)
                    self.ScanText.setTextBackgroundColor(White)
                    c = 0
                else:
                    self.ScanText.append(text)
                    c += 1
        return

###############################################################################################

    def WriteBigEndian(self, key):
        """
        Add data found in file scan to Big Endian tab
        """
        Yellow = QtGui.QColor(255,255,0)
        White = QtGui.QColor(255,255,255)
        Green = QtGui.QColor(0,100,0)
        Blue = QtGui.QColor(0,0,153)
        LightBlue = QtGui.QColor(153,204,255)
        Black = QtGui.QColor(0,0,0)

        tab_header = ("Trace Name\n"
                      "Start Time \t\t\t End Time \t\t\t Blockettes")
        self.big_e_text.setTextBackgroundColor(Yellow)
        self.big_e_text.setText(tab_header)
        self.big_e_text.setTextBackgroundColor(White)

        keylist = []
        keylist = list(self.BigEndianDict.keys())
        keylist.sort()

        if not self.BigEndianDict:
            return
        c = 0
        if key == "All":
            for indexkey in list(self.BigEndianDict.keys()):
                text = "<" + indexkey + ">"
                self.big_e_text.setTextColor(Blue)
                self.big_e_text.append(text)
                for directory in list(self.BigEndianDict[indexkey].keys()):
                    text2 = "  <" + directory + ">"
                    self.big_e_text.setTextColor(Green)
                    self.big_e_text.append(text2)
                    for trace in self.BigEndianDict[indexkey][directory]:
                        text = 4 * SPACE + str(trace[0])
                        templist = list(map(str, (trace[1:])))
                        text = text + "\n" + 6 * SPACE + "\t\t".join(templist)
                        self.big_e_text.setTextColor(Black)
                        if c:
                            self.big_e_text.setTextBackgroundColor(LightBlue)
                            self.big_e_text.append(text)
                            self.big_e_text.setTextBackgroundColor(White)
                            c = 0
                        else:
                            self.big_e_text.append(text)
                            c = 1
        else:
            numtrace = 0
            text = "<" + str(key) + ">"
            self.big_e_text.setTextColor(Blue)
            self.big_e_text.append(text)
            for directory in list(self.BigEndianDict[key].keys()):
                text2 = "  <" + directory + ">"
                self.big_e_text.setTextColor(Green)
                self.big_e_text.append(text2)
                for trace in self.BigEndianDict[key][directory]:
                    text = 4 * SPACE + str(trace[0])
                    templist = list(map(str, (trace[1:])))
                    text = text + "\n" + 6 * SPACE + "\t\t".join(templist)
                    self.big_e_text.setTextColor(Black)
                    if c:
                        self.big_e_text.setTextBackgroundColor(LightBlue)
                        self.big_e_text.append(text)
                        self.big_e_text.setTextBackgroundColor(White)
                        c = 0
                    else:
                        self.big_e_text.append(text)
                        c = 1
                    numtrace += 1

            if numtrace == 1:
                text = "Displaying " + str(numtrace) + " trace."
            else:
                text = "Displaying " + str(numtrace) + " trace."
            self.files_alert.setText(text)
            self.files_alert.setAlignment(QtGui.Qt.AlignCenter)
        return

###############################################################################################

    def WriteLittleEndian(self, key):
        """
        Add data found in file scan to Little Endian tab
        """
        Yellow = QtGui.QColor(255,255,0)
        White = QtGui.QColor(255,255,255)
        Green = QtGui.QColor(0,100,0)
        Blue = QtGui.QColor(0,0,153)
        LightBlue = QtGui.QColor(153,204,255)
        Black = QtGui.QColor(0,0,0)

        tab_header = ("Trace Name\n"
                      "Start Time \t\t\t End Time \t\t\t Blockettes")
    
        self.little_e_text.setTextBackgroundColor(Yellow)
        self.little_e_text.setText(tab_header)
        self.little_e_text.setTextBackgroundColor(White)

        keylist = []
        keylist = list(self.LittleEndianDict.keys())
        keylist.sort()

        if not self.LittleEndianDict:
            return
        c = 0
        if key == "All":
            for indexkey in list(self.LittleEndianDict.keys()):
                text = "<" + indexkey + ">"
                self.little_e_text.setTextColor(Blue)
                self.little_e_text.append(text)
                for directory in list(self.LittleEndianDict[indexkey].keys()):
                    text2 = "  <" + directory + ">"
                    self.little_e_text.setTextColor(Green)
                    self.little_e_text.append(text2)
                    for trace in self.LittleEndianDict[indexkey][directory]:
                        text = 4 * SPACE + str(trace[0])
                        templist = list(map(str, (trace[1:])))
                        text = text + "\n" + 6 * SPACE + "\t\t".join(templist)
                        self.little_e_text.setTextColor(Black)
                        if c:
                            self.little_e_text.setTextBackgroundColor(LightBlue)
                            self.little_e_text.append(text + "\n")
                            self.little_e_text.setTextBackgroundColor(White)
                            c = 0
                        else:
                            self.little_e_text.append(text + "\n")
                            c = 1
        else:
            numtrace = 0
            text = "<" + key + ">"
            self.little_e_text.setTextColor(Blue)
            self.little_e_text.append(text)
            for directory in list(self.LittleEndianDict[key].keys()):
                text2 = "  <" + directory + ">"
                self.little_e_text.setTextColor(Green)
                self.little_e_text.append(text2)
                for trace in self.LittleEndianDict[key][directory]:
                    text = 4 * SPACE + str(trace[0])
                    templist = list(map(str, (trace[1:])))
                    text = text + "\n" + 6 * SPACE + "\t\t".join(templist)
                    self.little_e_text.setTextColor(Black)
                    if c:
                        self.little_e_text.setTextBackgroundColor(LightBlue)
                        self.little_e_text.append(text + "\n")
                        self.little_e_text.setTextBackgroundColor(White)
                        c = 0
                    else:
                        self.little_e_text.append(text + "\n")
                        c = 1
                    numtrace += 1
            text = "Displaying " + str(numtrace) + " traces"
            self.files_alert.setText(text)
            self.files_alert.setAlignment(QtGui.Qt.AlignCenter)
        return

###############################################################################################

    def DisplayError(self):
        """
        Filters Error entries
        """
        self.error_text.clear()
        if self.ErrorType == 0:
            textlist = self.ErrorAll
        if self.ErrorType == 1:
            textlist = self.ErrorRW
        if self.ErrorType == 2:
            textlist = self.ErrorSize
        if self.ErrorType == 3:
            textlist = self.ErrorEndian
        if self.ErrorType == 4:
            textlist = self.ErrorUnique
        for (textcmd, color) in textlist:
            self.error_text.append(textcmd, color)
        return

###############################################################################################

    def UpdateInfoBar(self, numfiles, numerrors):
        """
        Updates tab names with number of files/errors found and updates the InfoBar
        """
        # text info bar
        if numerrors == 0:
            self.files_alert.setStyleSheet("background-color:green")
        else:
            self.files_alert.setStyleSheet("background-color:red")
        self.files_alert.setText(str(numfiles) + " unique mseed files found. *** " + str(numerrors) + " scan errors.")
        self.files_alert.setAlignment(QtGui.Qt.AlignCenter)

        # tab info
        self.tabWidget.setTabText(1, "Big Endian (" + str(self.NumBigFiles) + ")")
        self.tabWidget.setTabText(2, "Little Endian (" + str(self.NumLittleFiles) + ")")
        self.tabWidget.setTabText(3, "Errors (" + str(numerrors) + ")")

###############################################################################################
  
    def addTextInfoBar(self, str='', bg='yellow'):
        """
        Adds Text and color to InfoBar
        """
        # only bell if error, e.g. color red
        # don't ring bell if clearing text or progress info (e.g. color ==
        # yellow/lightblue)
        if BATCHMODE:
            if str:
                if (string.count(str, 'Examining') or
                        string.count(str, 'Modifying') or
                        string.count(str, 'Correcting') or
                        string.count(str, 'Remaining')):
                    print("\r" + str, end=' ')
                    sys.stdout.flush()
                else:
                    print(str)
        else:
            if bg != "yellow" and bg != "lightblue":
                self.PlaySound()
            text = str
            self.files_alert.setText(text)
            self.files_alert.setAlignment(QtGui.Qt.AlignCenter)
            self.files_alert.setStyleSheet("background-color:" + bg)

###############################################################################################
    
    def big_e_dropmenu_selections(self, event):
        """
        Action function for Big Endian drop menu
        """ 
        index = self.big_e_dropmenu.currentText() 
        if index == "All":
            self.WriteBigEndian("All")
        else: 
            self.WriteBigEndian(index)

###############################################################################################

    def little_e_dropmenu_selections(self, event):
        """
        Action function for Little Endian drop menu
        """ 
        index = self.little_e_dropmenu.currentText() 
        if index == "All":
            self.WriteLittleEndian("All")
        else:
            self.WriteLittleEndian(index)

###############################################################################################

    def clicked_save_scan(self):
        """
        Action function for clicking Save Scan. Creates default file name 
        and calls window pop up function
        """
        (year, month, day, hour, minute, second, weekday,
         yearday, daylight) = time.localtime(time.time())
        
        templist = list(map(str, (year, yearday, hour, minute)))
        now = ".".join(templist)

        self.Savefile = ("ckMseed." + now)

        self.create_save_window(self.Savefile)


###############################################################################################

    def create_save_window(self, Savefile):
        """
        Init window to save scan
        """
        # Window settings
        self.save_window = QtWidgets.QWidget()
        self.save_window.setFixedSize(400,200)
        self.save_window.setWindowTitle("Save Scan File as:")

        # Save file name
        self.save_file_name = QtWidgets.QTextEdit()
        self.save_file_name.setText(Savefile)

        # Initialize buttons
        self.save_btn = QtWidgets.QPushButton("Save")
        self.save_btn.setStyleSheet("""
                                    QPushButton{
                                    background-color:rgb(98, 160, 234);;
                                    }
                                    QPushButton::hover
                                    {
                                    background-color:green;
                                    }
                                    """)
        self.save_btn.clicked.connect(lambda: self.WriteFile())
        
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        self.cancel_btn.setStyleSheet("QPushButton::hover"
                                    "{"
                                    "background-color:red;"
                                    "}")
        self.cancel_btn.clicked.connect(lambda: self.close_save_window())

        # Initialize checkboxes
        self.SaveAll = QtWidgets.QCheckBox()
        self.SaveAll.setText("All")
        self.SaveAll.toggled.connect(self.save_all_toggled)
        self.SaveInfo = QtWidgets.QCheckBox()
        self.SaveInfo.setText("Info")
        self.SaveErrors = QtWidgets.QCheckBox()
        self.SaveErrors.setText("Error")
        self.SaveErrorTraces = QtWidgets.QCheckBox()
        self.SaveErrorTraces.setText("Error Traces")

        # Initialize layouts
        file_layout = QtWidgets.QVBoxLayout()
        file_layout.addWidget(self.save_file_name)


        check_box_layout = QtWidgets.QHBoxLayout()
        check_box_layout.addWidget(self.SaveAll)
        check_box_layout.addWidget(self.SaveInfo)
        check_box_layout.addWidget(self.SaveErrors)
        check_box_layout.addWidget(self.SaveErrorTraces)
        check_box_layout.setAlignment(QtCore.Qt.AlignCenter)

        buttons_layout = QtWidgets.QVBoxLayout()
        buttons_layout.addWidget(self.save_btn)
        buttons_layout.addWidget(self.cancel_btn)
        buttons_layout.setAlignment(QtCore.Qt.AlignBottom)

        total_layout = QtWidgets.QVBoxLayout()
        total_layout.addLayout(file_layout)
        total_layout.addLayout(check_box_layout)
        total_layout.addLayout(buttons_layout)

        self.save_window.setLayout(total_layout)

        self.save_window.show()

###############################################################################################

    def close_save_window(self):
        """
        Action function to close the save window
        """
        self.save_window.close()
        self.addTextInfoBar()

###############################################################################################

    def save_all_toggled(self):
        """
        If "All" is selected, check all other boxes.
        If "All" is unselected, uncheck all other boxes.
        """
        if self.SaveAll.isChecked():
            self.SaveInfo.setChecked(True)
            self.SaveErrors.setChecked(True)
            self.SaveErrorTraces.setChecked(True)
        else:
            self.SaveInfo.setChecked(False)
            self.SaveErrors.setChecked(False)
            self.SaveErrorTraces.setChecked(False)

###############################################################################################

    def create_overwrite_window(self):
        """
        If file already exists ask to overwrite
        """
        self.result = ""
        self.warning = QtWidgets.QMessageBox()
        self.warning.setIcon(QtWidgets.QMessageBox.Warning)
        self.warning.setWindowTitle("Warning")
        self.warning.setText("The file exists!")

        overwrite = self.warning.addButton("Overwrite", QtWidgets.QMessageBox.ActionRole)
        cancel = self.warning.addButton("Cancel", QtWidgets.QMessageBox.ActionRole)
        
        self.warning.exec_()

        if self.warning.clickedButton() == overwrite:
            self.result = "overwrite"
            return
        elif self.warning.clickedButton() == cancel:
            self.result = "cancel"
            self.addTextInfoBar()
            return
    
############################################################################################### 
    
    def WriteFile(self):
        """
        Write File to disk
        """
        divider = "///////////////////"

        file = self.save_file_name.toPlainText()
        if os.path.isfile(file):
            self.create_overwrite_window()
        else:
            self.result = "overwrite"

        if self.result == "overwrite":

            if (not self.SaveAll.isChecked()) and (not self.SaveInfo.isChecked()) and (not self.SaveErrors.isChecked()) and (not self.SaveErrorTraces.isChecked()):
                oops = "Nothing to save"
                self.addTextInfoBar(oops, 'orange')
                return

            try:
                outfile = open(file, "w")

                if (self.SaveAll.isChecked()) or (self.SaveInfo.isChecked()):
                    if self.StatChanLocDict:
                            outfile.write("%s\n" % divider)
                            outfile.write("%s\n" % "Trace Information")
                            outfile.write("%s\n" % divider)
                            keylist = []
                            keylist = list(self.StatChanLocDict.keys())
                            keylist.sort()

                            if keylist:
                                for key in keylist:
                                    numspace = 24 - len(key)
                                    if not self.ScanType:
                                        text = key
                                    else:
                                        start = time.strftime('%Y:%j:%H:%M:%S',
                                                            time.gmtime(self.StatChanLocDict[key][0][0]))  # noqa: E501
                                        end = time.strftime('%Y:%j:%H:%M:%S',
                                                            time.gmtime(self.StatChanLocDict[key][0][1]))  # noqa: E501
                                        templist = list(map(str, (start, end)))
                                        text = key + numspace * \
                                            SPACE + "   ".join(templist)

                                    outfile.write("%s\n" % text)
                    else:
                        outfile.close()
                        os.unlink(file)
                        oops = "Nothing to save"
                        self.addTextInfoBar(oops, 'orange')
                        return

                if self.SaveErrors.isChecked():
                    if not self.SaveInfo.isChecked():
                        outfile.write("%s\n" % divider)
                    else:
                        outfile.write("\n\n%s\n" % divider)
                    outfile.write("%s\n" % "All Errors")
                    outfile.write("%s\n" % divider)
                    if self.ErrorAll:
                        for (text, color) in self.ErrorAll:
                            outfile.write("%s" % text)
                    else:
                        outfile.write("%s\n" % "NONE")

                if self.SaveErrorTraces.isChecked():
                    if not self.SaveInfo.isChecked() or not self.SaveErrors.isChecked():
                            outfile.write("%s\n" % divider)
                    else:
                        outfile.write("\n\n%s\n" % divider)
                    outfile.write("%s\n" % "Traces with Errors")
                    outfile.write("%s\n" % divider)
                    if self.ErrorDict:
                        for indexkey in list(self.ErrorDict.keys()):
                            text = "<" + indexkey + ">\n"
                            outfile.write("%s" % text)
                            for directory in list(self.ErrorDict[indexkey].keys()):  # noqa: E501
                                text2 = "   <" + directory + ">\n"
                                outfile.write("%s" % text2)
                                for trace in self.ErrorDict[indexkey][directory]:  # noqa: E501
                                    text3 = "\t" + trace + "\n"
                                    outfile.write("%s" % text3)
                    else:
                        outfile.write("%s\n" % "NONE")

                
                outfile.close()

            except IOError as e:
                err = "Can't Create %s" % file, e
                self.addTextInfoBar(err, 'red')
                return

        self.save_window.close()
        
###############################################################################################
            
if __name__ == "__main__":
    # get commandline args
    try:
        opts, pargs = getopt(sys.argv[1:], 'hsvV#d:')
    except Exception:
        err = "\nERROR: Invalid command line usage\n"
        Usage(err)
        sys.exit(1)
    else:
        Dataloc = ""
        RunFlag = ""
        BatchFile = 1
        for flag, arg in opts:
            if flag == "-h":
                Usage()
                sys.exit(0)
            if flag == "-d":
                Dataloc = arg
            if flag == "-s":
                RunFlag = "Simple"
            if flag == "-v":
                RunFlag = "Verbose"
            if flag == "-V":
                RunFlag = "Very Verbose"
            if flag == "-#":
                print(VERSION)
                sys.exit(0)
    print("\n", os.path.basename(sys.argv[0]), VERSION)

    app = QtWidgets.QApplication(sys.argv)
    Window = MainWindow(BatchFile=BatchFile, Dataloc=Dataloc, RunFlag=RunFlag)
    sys.exit(app.exec_())